from __future__ import unicode_literals
from prompt_toolkit.application.current import get_app
from prompt_toolkit.eventloop.context import TaskLocal, TaskLocalNotSetError
from prompt_toolkit.utils import is_windows
from .base import Input
import sys

__all__ = [
    'create_input',
    'create_pipe_input',
    'get_default_input',
    'set_default_input',
]


def create_input(stdin=None):
    stdin = stdin or sys.stdin

    if is_windows():
        from .win32 import Win32Input
        return Win32Input(stdin)
    else:
        from .vt100 import Vt100Input
        return Vt100Input(stdin)


def create_pipe_input():
    """
    Create an input pipe.
    This is mostly useful for unit testing.
    """
    if is_windows():
        from .win32_pipe import Win32PipeInput
        return Win32PipeInput()
    else:
        from .posix_pipe import PosixPipeInput
        return PosixPipeInput()


_default_input = TaskLocal()


def get_default_input():
    """
    Get the input class to be used by default.

    Called when creating a new Application(), when no `Input` has been passed.
    """
    # Other create/return the default input.
    try:
        value = _default_input.get()
    except TaskLocalNotSetError:
        # If an application is already running, take the input from there.
        # (This is important for the "ENTER for continue" prompts after
        # executing system commands and displaying readline-style completions.)
        app = get_app(return_none=True)
        if app:
            return app.input

        return create_input()
    else:
        return value


def set_default_input(input):
    """
    Set the default `Output` class.

    (Used for instance, for the telnet submodule.)
    """
    assert isinstance(input, Input)
    _default_input.set(input)



"""
Task local storage for coroutines.

Example::

    # Create a new task local.
    my_local = TaskLocal()

    # Set/retrieve/erase:
    my_local.set(value)

    value = my_local.get()

    my_local.delete()

A new scope can be created as follows::

    with context():
        ...

Within this scope, a new value can be assigned, which is only visible within
the scope. The scope as passed along when code is sent to an executor and back.
"""
from __future__ import unicode_literals
from threading import local
from functools import wraps

__all__ = [
    'context',
    'get_context_id',
    'wrap_in_current_context',
    'TaskLocal',
    'TaskLocalNotSetError',
]


_storage = local()
_last_context_id = 0


def get_context_id():
    " Return the current context ID or None. "
    try:
        return _storage.context_id
    except AttributeError:
        return 0  # Default context.


class context(object):
    """
    Context manager that activates a new scope.
    """
    def __init__(self, context_id=None):
        global _last_context_id

        if context_id is not None:
            self.id = context_id
        else:
            _last_context_id += 1
            self.id = _last_context_id

    def __enter__(self):
        try:
            self._previous_id = _storage.context_id
        except AttributeError:
            self._previous_id = None

        _storage.context_id = self.id
        return self.id

    def __exit__(self, *a):
        if self._previous_id is None:
            del _storage.context_id
        else:
            _storage.context_id = self._previous_id


class TaskLocal(object):
    """
    Like a thread local, but tied to the current task.
    """
    def __init__(self):
        self._storage = {}

    def get(self):
        try:
            ctx = get_context_id()
            return self._storage[ctx]
        except KeyError:
            raise TaskLocalNotSetError

    def set(self, value):
        ctx = get_context_id()
        self._storage[ctx] = value

    def delete(self):
        ctx = get_context_id()
        try:
            del self._storage[ctx]
        except KeyError:
            pass


def wrap_in_current_context(func):
    """
    Decorator that takes a function, and ensures that when it's called, the
    current context will apply.
    """
    assert callable(func)
    ctx_id = get_context_id()

    @wraps(func)
    def new_func(*a, **kw):
        with context(ctx_id):
            return func(*a, **kw)
    return new_func


class TaskLocalNotSetError(Exception):
    pass



from __future__ import unicode_literals

__all__ = [
    'explode_text_fragments',
]


class _ExplodedList(list):
    """
    Wrapper around a list, that marks it as 'exploded'.

    As soon as items are added or the list is extended, the new items are
    automatically exploded as well.
    """
    def __init__(self, *a, **kw):
        super(_ExplodedList, self).__init__(*a, **kw)
        self.exploded = True

    def append(self, item):
        self.extend([item])

    def extend(self, lst):
        super(_ExplodedList, self).extend(explode_text_fragments(lst))

    def insert(self, index, item):
        raise NotImplementedError  # TODO

    # TODO: When creating a copy() or [:], return also an _ExplodedList.

    def __setitem__(self, index, value):
        """
        Ensure that when `(style_str, 'long string')` is set, the string will be
        exploded.
        """
        if not isinstance(index, slice):
            index = slice(index, index + 1)
        value = explode_text_fragments([value])
        super(_ExplodedList, self).__setitem__(index, value)


def explode_text_fragments(fragments):
    """
    Turn a list of (style_str, text) tuples into another list where each string is
    exactly one character.

    It should be fine to call this function several times. Calling this on a
    list that is already exploded, is a null operation.

    :param fragments: List of (style, text) tuples.
    """
    # When the fragments is already exploded, don't explode again.
    if getattr(fragments, 'exploded', False):
        return fragments

    result = []

    for style, string in fragments:
        for c in string:
            result.append((style, c))

    return _ExplodedList(result)



"""
Key bindings for auto suggestion (for fish-style auto suggestion).
"""
from __future__ import unicode_literals
import re
from prompt_toolkit.application.current import get_app
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.filters import Condition, emacs_mode

__all__ = [
    'load_auto_suggest_bindings',
]


def load_auto_suggest_bindings():
    """
    Key bindings for accepting auto suggestion text.

    (This has to come after the Vi bindings, because they also have an
    implementation for the "right arrow", but we really want the suggestion
    binding when a suggestion is available.)
    """
    key_bindings = KeyBindings()
    handle = key_bindings.add

    @Condition
    def suggestion_available():
        app = get_app()
        return (app.current_buffer.suggestion is not None and
                app.current_buffer.document.is_cursor_at_the_end)

    @handle('c-f', filter=suggestion_available)
    @handle('c-e', filter=suggestion_available)
    @handle('right', filter=suggestion_available)
    def _(event):
        " Accept suggestion. "
        b = event.current_buffer
        suggestion = b.suggestion

        if suggestion:
            b.insert_text(suggestion.text)

    @handle('escape', 'f', filter=suggestion_available & emacs_mode)
    def _(event):
        " Fill partial suggestion. "
        b = event.current_buffer
        suggestion = b.suggestion

        if suggestion:
            t = re.split(r'(\S+\s+)', suggestion.text)
            b.insert_text(next(x for x in t if x))

    return key_bindings



"""
Utilities for manipulating formatted text.

When ``to_formatted_text`` has been called, we get a list of ``(style, text)``
tuples. This file contains functions for manipulating such a list.
"""
from __future__ import unicode_literals
from prompt_toolkit.utils import get_cwidth

__all__ = [
    'fragment_list_len',
    'fragment_list_width',
    'fragment_list_to_text',
    'split_lines',
]


def fragment_list_len(fragments):
    """
    Return the amount of characters in this text fragment list.

    :param fragments: List of ``(style_str, text)`` or
        ``(style_str, text, mouse_handler)`` tuples.
    """
    ZeroWidthEscape = '[ZeroWidthEscape]'
    return sum(len(item[1]) for item in fragments if ZeroWidthEscape not in item[0])


def fragment_list_width(fragments):
    """
    Return the character width of this text fragment list.
    (Take double width characters into account.)

    :param fragments: List of ``(style_str, text)`` or
        ``(style_str, text, mouse_handler)`` tuples.
    """
    ZeroWidthEscape = '[ZeroWidthEscape]'
    return sum(get_cwidth(c) for item in fragments for c in item[1] if ZeroWidthEscape not in item[0])


def fragment_list_to_text(fragments):
    """
    Concatenate all the text parts again.

    :param fragments: List of ``(style_str, text)`` or
        ``(style_str, text, mouse_handler)`` tuples.
    """
    ZeroWidthEscape = '[ZeroWidthEscape]'
    return ''.join(item[1] for item in fragments if ZeroWidthEscape not in item[0])


def split_lines(fragments):
    """
    Take a single list of (style_str, text) tuples and yield one such list for each
    line. Just like str.split, this will yield at least one item.

    :param fragments: List of (style_str, text) or (style_str, text, mouse_handler)
                      tuples.
    """
    line = []

    for item in fragments:
        # For (style_str, text) tuples.
        if len(item) == 2:
            style, string = item
            parts = string.split('\n')

            for part in parts[:-1]:
                if part:
                    line.append((style, part))
                yield line
                line = []

            line.append((style, parts[-1]))
                # Note that parts[-1] can be empty, and that's fine. It happens
                # in the case of [('[SetCursorPosition]', '')].

        # For (style_str, text, mouse_handler) tuples.
        #     I know, partly copy/paste, but understandable and more efficient
        #     than many tests.
        else:
            style, string, mouse_handler = item
            parts = string.split('\n')

            for part in parts[:-1]:
                if part:
                    line.append((style, part, mouse_handler))
                yield line
                line = []

            line.append((style, parts[-1], mouse_handler))

    # Always yield the last line, even when this is an empty line. This ensures
    # that when `fragments` ends with a newline character, an additional empty
    # line is yielded. (Otherwise, there's no way to differentiate between the
    # cases where `fragments` does and doesn't end with a newline.)
    yield line



"""
Styling for prompt_toolkit applications.
"""
from __future__ import unicode_literals, absolute_import

from .base import Attrs, DEFAULT_ATTRS, ANSI_COLOR_NAMES, BaseStyle, DummyStyle, DynamicStyle
from .defaults import default_ui_style, default_pygments_style
from .style import Style, Priority, merge_styles
from .pygments import style_from_pygments_cls, style_from_pygments_dict, pygments_token_to_classname
from .named_colors import NAMED_COLORS


__all__ = [
    # Base.
    'Attrs',
    'DEFAULT_ATTRS',
    'ANSI_COLOR_NAMES',
    'BaseStyle',
    'DummyStyle',
    'DynamicStyle',

    # Defaults.
    'default_ui_style',
    'default_pygments_style',

    # Style.
    'Style',
    'Priority',
    'merge_styles',

    # Pygments.
    'style_from_pygments_cls',
    'style_from_pygments_dict',
    'pygments_token_to_classname',

    # Named colors.
    'NAMED_COLORS',
]



"""
Filters decide whether something is active or not (they decide about a boolean
state). This is used to enable/disable features, like key bindings, parts of
the layout and other stuff. For instance, we could have a `HasSearch` filter
attached to some part of the layout, in order to show that part of the user
interface only while the user is searching.

Filters are made to avoid having to attach callbacks to all event in order to
propagate state. However, they are lazy, they don't automatically propagate the
state of what they are observing. Only when a filter is called (it's actually a
callable), it will calculate its value. So, its not really reactive
programming, but it's made to fit for this framework.

Filters can be chained using ``&`` and ``|`` operations, and inverted using the
``~`` operator, for instance::

    filter = has_focus('default') & ~ has_selection
"""
from __future__ import unicode_literals

from .base import Filter, Never, Always, Condition
from .app import *
from .utils import to_filter

# For backwards-compatibility. Keep this import.
from .cli import *
